class FastFsException(Exception):
    """Base exception class for fastfs library."""

    def __init__(self, message: str):
        self.message = message
        super().__init__(self.message)


class UnsupportedFileType(FastFsException):
    """Raised when an unsupported file type is used."""

    def __init__(self, file_type: str, supported_types: str = ''):

        if len(supported_types) > 0:
            self.message = f"The file type {file_type} is not supported. Supported types: {supported_types}"
        else:
            self.message = f"The file type {file_type} is not supported."

        super().__init__(self.message)


class MissingDependencyError(FastFsException):
    """Raised when a required dependency is not installed."""

    def __init__(self, dependency_name: str):
        self.message = f"The {dependency_name} package is not installed. Please install this package to use the requested feature."
        super().__init__(self.message)


class InvalidFileDataError(FastFsException):
    """Raised when the provided data does not match the expected format for the file type."""

    def __init__(self, explanation: str):
        self.message = f"The provided data does not match the expected format for the file type. {explanation}"
        super().__init__(self.message)


class CorruptFileError(FastFsException):
    def __init__(self, explanation: str):
        self.message = f"The file could not be read/written. {explanation}"
        super().__init__(self.message)


class FileNotFound(FastFsException):
    """Raised when a specified file is not found."""

    def __init__(self, file_name: str):
        self.message = f"The file {file_name} does not exist."
        super().__init__(self.message)


class DirectoryNotFound(FastFsException):
    """Raised when a specified file is not found."""

    def __init__(self, file_name: str):
        self.message = f"The file {file_name} does not exist."
        super().__init__(self.message)


class FileWriteError(FastFsException):
    """Raised when there is an error writing to a file."""

    def __init__(self):
        self.message = "There was an error writing to the file."
        super().__init__(self.message)


class FileReadError(FastFsException):
    """Raised when there is an error reading from a file."""

    def __init__(self):
        self.message = "There was an error reading from the file."
        super().__init__(self.message)


class BulkReadDirectoryError(FastFsException):
    """Raised when there is an error from the bulk_read_directory function in AbstractFileManager."""

    def __init__(self, explanation: str):
        self.message = f"There was an error during bulk directory read. {explanation}"
        super().__init__(self.message)



import os
import json
import pickle
import unittest
from unittest.mock import patch, ANY
from fastfs import write_pickle
from fastfs.utils import create_fs, get_fs_directory, touch_directory

from fastfs.global_instance import fast_file_manager

import configparser

import shutil


class FastFSDirectoryTests(unittest.TestCase):
    """Tests the functionality of the .fastfs relative file system"""

    def setUp(self):
        self.fast_fs_dir = "test_fastfs_files"
        # Check that environment is clean before starting
        self.assertFalse(os.path.exists(self.fast_fs_dir))
        self.assertFalse(os.path.exists('.fastfs'))
        self.assertIsNone(fast_file_manager._local_fs)
        self.assertFalse(fast_file_manager._fs_active)

        # Create fs directory
        create_fs(self.fast_fs_dir, active=True)

    def _reset_fs_directory(self):
        if os.path.exists(self.fast_fs_dir):
            shutil.rmtree(self.fast_fs_dir)

        if os.path.exists('.fastfs'):
            os.remove('.fastfs')

        fast_file_manager._local_fs = None
        fast_file_manager._fs_active = False

    def tearDown(self):
        self._reset_fs_directory()

    def _test_create_fs(self, test_active):

        # Since it's created by default in setUp()
        self._reset_fs_directory()

        create_fs(self.fast_fs_dir, test_active)

        # Make sure files got created properly
        self.assertTrue(os.path.exists(self.fast_fs_dir),
                        f'{self.fast_fs_dir} does not exist')
        self.assertTrue(os.path.exists('.fastfs'), '.fastfs does not exist')

        # Create a ConfigParser object
        config = configparser.ConfigParser()

        # Read from a configuration file
        config.read('.fastfs')

        # Test that all of the values are in the config
        self.assertIn('DEFAULT', config,
                      'DEFAULT section missing from .fastfs')
        self.assertIn('directory', config['DEFAULT'],
                      'directory missing from DEFAULT section')
        self.assertIn('active', config['DEFAULT'],
                      'active missing from DEFAULT section')

        # Test that directory and active are proper values
        self.assertEqual(config['DEFAULT']['directory'], self.fast_fs_dir)

        active = config.getboolean('DEFAULT', 'active')

        self.assertEqual(active, test_active,
                         '.fastfs active is not the correct value!')

    def test_create_fs_active(self):

        self._test_create_fs(True)

    def test_create_fs_inactive(self):

        self._test_create_fs(False)

    def test_fs_relative_file_path(self):

        write_pickle('test.pkl', [])

        path = os.path.join(self.fast_fs_dir, 'test.pkl')

        self.assertTrue(path)

    def test_get_fs_directory(self):

        fs_directory = get_fs_directory()

        self.assertEqual(fs_directory, self.fast_fs_dir)

    def test_fs_relative_directory_path(self):
        test_directory = 'test_dir'

        touch_directory(test_directory)

        fs_path = os.path.join(self.fast_fs_dir, test_directory)

        # Make sure that the fastfs directory is being used
        self.assertTrue(os.path.exists(fs_path))
        self.assertFalse(os.path.exists(test_directory))

    def test_absolute_path(self):

        test_directory = 'test_dir'

        self.assertNotEqual(self.fast_fs_dir, test_directory)

        cwd = os.getcwd()

        full_path = os.path.join(cwd, test_directory)

        touch_directory(full_path)

        self.assertTrue(os.path.exists(full_path))

        # Make sure fastfs directory wasn't in use despite it being enabled because we're using an abs path
        self.assertFalse(os.path.exists(
            os.path.join(self.fast_fs_dir, test_directory)))

        # Delete the test directory
        shutil.rmtree(full_path)


if __name__ == "__main__":
    unittest.main()



from typing import Any, Union, List, Tuple, Dict

from fastfs.global_instance import fast_file_manager


def write_pickle(file_name: str, file_data: Any):
    """
    Writes data to a pickle file.

    Args:
        file_name: The name/path of the file to write the pickle data to.
        file_data: The data to write as a pickle object.
    """
    fast_file_manager.write_pickle(file_name, file_data)


def write_json(file_name: str, file_data: Any):
    """
    Writes data to a JSON file.

    Args:
        file_name: The name/path of the file to write the JSON data to.
        file_data: The data to write as a JSON object.
    """
    fast_file_manager.write_json(file_name, file_data)


def write_csv(file_name: str, file_data: Union[List[dict], List[list]], header: Union[None, list] = None):
    """
    Writes data to a CSV file.

    Args:
        file_name: The name/path of the file to write the CSV data to.
        file_data: The data to write as a list of dictionaries or a list of lists.
        header: An optional list of header values. If provided, this will be written as the first row in the CSV file.
    """
    fast_file_manager.write_csv(file_name, file_data, header=header)


def read_csv(file_name: str, return_list_of_dicts: bool = False) -> Union[Tuple[List[str], List[List[str]]], List[dict]]:
    """
    Reads data from a CSV file.

    Args:
        file_name: The name/path of the CSV file to read from.

    Returns:
        Union[Tuple[List[str], List[List[str]]], List[dict]]: The data read from the CSV file, either a tuple containing headers and rows 
        or a single list of dicts with the headers as the keys in the list.
    """

    return fast_file_manager.read_csv(file_name, return_list_of_dicts=return_list_of_dicts)


def write_file(file_name: str, file_data: Any):
    """
    Writes data to a file.

    Args:
        file_name: The name/path of the file to write the data to.
        file_data: The data to write to the file.
    """
    fast_file_manager.write_file(file_name, file_data)


def read_pickle(file_name: str) -> Any:
    """
    Reads data from a pickle file.

    Args:
        file_name: The name/path of the pickle file to read from.

    Returns:
        Any: The data read from the pickle file.
    """
    return fast_file_manager.read_pickle(file_name)


def read_json(file_name: str) -> Union[dict, list]:
    """
    Reads data from a JSON file.

    Args:
        file_name: The name/path of the JSON file to read from.

    Returns:
        Union[dict, list]: The data read from the JSON file, either a dictionary or a list.
    """
    return fast_file_manager.read_json(file_name)


def read_file(file_name: str) -> str:
    """
    Reads data from a file.

    Args:
        file_name: The name/path of the file to read from.

    Returns:
        str: The data read from the file.
    """
    return fast_file_manager.read_file(file_name)


def write_lines(file_name: str, lines: list):
    """
    Writes a list of lines to a file.

    Args:
        file_name: The name/path of the file to write the lines to.
        lines: A list of strings to write to the file, one string per line.
    """
    fast_file_manager.write_lines(file_name, lines)


def read_lines(file_name: str) -> List[str]:
    """
    Reads lines from a file.

    Args:
        file_name: The name/path of the file to read lines from.

    Returns:
        List[str]: A list of strings containing the lines read from the file.
    """
    return fast_file_manager.read_lines(file_name)


def write_ini(file_name: str, data: Union[Dict[Any, Dict[Any, Any]], Dict[Any, Any]]):
    """
    Writes data to an INI file.

    Args:
        file_name: The name/path of the file to write the data to.
        data: The data to write to the INI file. It can either be a dict or a dict of dicts.
        If data is only a dict, the data will be written to under the 'DEFAULT' section.
    """
    fast_file_manager.write_ini(file_name, data)


def read_ini(file_name: str) -> Union[Dict[Any, Dict[Any, Any]], Dict[Any, Any]]:
    """
    Reads data from an INI file.

    Args:
        file_name: The name/path of the INI file to read from.

    Returns:
        Union[Dict[Any, Dict[Any, Any]], Dict[Any, Any]]: The data read from the INI file. 
        If the INI file has multiple sections, it returns a dict of dicts. If it only has a default section, it returns a flat dict.
    """
    return fast_file_manager.read_ini(file_name)




from typing import Any


def path_replace(func):
    def wrapper(self, file_path, *args, **kwargs):
        file_path = self._path_replace(file_path)
        return func(self, file_path, *args, **kwargs)
    return wrapper


def safe_write(write_mode='w'):
    def decorator(func):
        def wrapper(self, file_name: str, file_data: Any, *args, **kwargs):
            self._safe_write_func(file_name, func, file_data,
                                  write_mode, *args, **kwargs)
        return wrapper
    return decorator


def safe_read(read_mode='r', context_manager=True):
    def decorator(func):
        def wrapper(self, file_name: str, *args, **kwargs):
            return self._safe_read_func(file_name, func, read_mode, context_manager, *args, **kwargs)
        return wrapper
    return decorator



import os
import unittest
from fastfs import write_pickle, read_pickle, write_json, read_json, write_lines, read_lines, write_csv, read_csv, write_file, read_file, write_ini, read_ini

import shutil

class FastFSTests(unittest.TestCase):

    def _get_path(self, file_name):

        return os.path.join(self.test_dir, file_name)

    def setUp(self):
        self.test_dir = "test_files"
        self.test_data = {"test": "FastFS Unit Test"}

         # Prepare the test directory
        if not os.path.exists(self.test_dir):
            os.mkdir(self.test_dir)

        self.pickle_path = self._get_path('test.pkl')
        self.json_path = self._get_path('test.json')
        self.csv_path = self._get_path('test.csv')
        self.file_path = self._get_path('test.txt')
        self.ini_path = self._get_path('test.ini')

    def tearDown(self):
    
        if os.path.exists(self.test_dir):
            shutil.rmtree(self.test_dir)


    # Tests for write_pickle & read_pickle methods
    def test_pickle_file_write_and_read(self):
        # Writing to pickle file
        write_pickle(self.pickle_path, self.test_data)

        # Verifying if file is created
        self.assertTrue(os.path.exists(self.pickle_path))

        # Reading from pickle file
        read_data = read_pickle(self.pickle_path)

        # Verification
        self.assertEqual(self.test_data, read_data)

    # Tests for write_json & read_json methods
    def test_json_file_write_and_read(self):
        # Writing to json file
        write_json(self.json_path, self.test_data)

        # Verifying if file is created
        self.assertTrue(os.path.exists(self.json_path))

        # Reading from json file
        read_data = read_json(self.json_path)

        # Verification
        self.assertEqual(self.test_data, read_data)

    # Tests for write_lines & read_lines methods
    def test_lines_write_and_read(self):
        lines = ['Line 1', 'Line 2', 'Line 3']

        # Writing to file
        write_lines(self.file_path, lines)

        # Reading from file
        read_lines_data = read_lines(self.file_path)

        # Verification
        self.assertEqual(lines, read_lines_data)

    def test_csv_write_and_read_with_headers(self):

        # Test data for writing and reading
        headers = ['Name', 'Age', 'Occupation']
        data_lists = [['Alice', '30', 'Engineer'],
                      ['Bob', '40', 'Doctor']]

        # Write and read as list of lists
        write_csv(self.csv_path, data_lists, header=headers)
        read_headers, read_data = read_csv(self.csv_path, return_list_of_dicts=False)
        self.assertEqual(headers, read_headers)
        self.assertEqual(data_lists, read_data)

    def test_csv_write_and_read_no_headers(self):

        # Test data for writing and reading
        data_dicts = [{'Name': 'Alice', 'Age': '30', 'Occupation': 'Engineer'},
                      {'Name': 'Bob', 'Age': '40', 'Occupation': 'Doctor'}]
   

        # Write and read as list of dicts
        write_csv(self.csv_path, data_dicts)
        read_data = read_csv(self.csv_path, return_list_of_dicts=True)
        self.assertEqual(data_dicts, read_data)

    def test_file_write_and_read(self):
        # Test data for writing and reading
        data = 'This is some test data.'

        # Write data to file
        write_file(self.file_path, data)

        # Read data back from file
        read_data = read_file(self.file_path)

        self.assertEqual(data, read_data)

    def test_ini_write_and_read(self):

        data = {'test': 'value'}

        write_ini(self.ini_path, data)

        read_data = read_ini(self.ini_path)

        self.assertEqual(data, read_data)


if __name__ == "__main__":

    unittest.main()



# Utils
from typing import Callable, Any, List, Union
from fastfs.global_instance import fast_file_manager
from fastfs.data_types import FileTypes


def get_fs_directory(absolute_path: bool = False) -> str:
    """
    Returns the fastfs directory.

    Args:
        absolute_path: If True, returns the absolute path of the directory.

    Returns:
        str: The fastfs directory path.
    """
    return fast_file_manager.get_fs_directory(absolute_path=absolute_path)


def ls(directory_name: str) -> List[str]:
    """
    Lists all files in the given directory.

    Args:
        directory_name: The name/path of the directory.

    Returns:
        List[str]: A list of file names in the directory.
    """
    return fast_file_manager.ls(directory_name)


def sorted_ls(directory_name: str, sort_by: Callable[[str], Any], reverse: bool = False) -> List[str]:
    """
    Lists all files in a directory, sorted based on the given sorting function.

    Args:
        directory_name: The name/path of the directory to list files from.
        sort_by: A callable function that takes a file name as input and returns a sorting key.
        reverse: If True, sorts the files in reverse order.

    Returns:
        List[str]: A list of sorted file names in the directory.
    """
    return fast_file_manager.sorted_ls(directory_name, sort_by, reverse=reverse)


def create_fs(directory_name: str = 'files', active: bool = True):
    """
    Creates a new fastfs directory.

    Args:
        directory_name: The name/path of the new directory.
        active: If True, sets the new directory as the active fastfs directory.
    """
    fast_file_manager.create_fs(directory_name, active=active)


def file_exists(file_name: str) -> bool:
    """
    Checks if a file with the given name exists.

    Args:
        file_name: The name/path of the file to check.

    Returns:
        bool: True if the file exists, otherwise False.
    """
    return fast_file_manager.file_exists(file_name)


def touch_file(file_name: str):
    """
    Creates a new, empty file with the given name.

    Args:
        file_name: The name/path of the file to create.
    """
    return fast_file_manager.touch_file(file_name)


def delete_file(file_name: str):
    """
    Deletes the file with the given name.

    Args:
        file_name: The name/path of the file to delete.

    Raises:
        FileNotFoundError: If the file does not exist.
    """
    fast_file_manager.delete_file(file_name)


def delete_directory(directory_name: str):
    """
    Deletes the directory with the given name.

    Args:
        directory_name: The name/path of the directory to delete.

    Raises:
        FileNotFoundError: If the directory does not exist.
    """
    fast_file_manager.delete_directory(directory_name)


def touch_directory(directory_name: str):
    """
    Creates a new directory with the given name.

    Args:
        directory_name: The name/path of the directory to create.
    """
    fast_file_manager.touch_directory(directory_name)


def get_directory_info(directory_name: str) -> dict:
    """
    Returns information about the given directory.

    Args:
        directory_name: The name/path of the directory.

    Returns:
        dict: A dictionary containing information about the directory, including the absolute path, number of files,
              creation time, modification time, and total size.
    """
    return fast_file_manager.get_directory_info(directory_name)


def get_file_info(file_name: str) -> dict:
    """
    Returns information about the given file.

    Args:
        file_name: The name/path of the file.

    Returns:
        dict: A dictionary containing information about the file, including the absolute path, creation time,
              modification time, size, and extension.
    """
    return fast_file_manager.get_file_info(file_name)


def bulk_write_directory(directory_name: str, file_data_ls: List[Any], data_type: Union[FileTypes, str],
                         file_prefix: Union[None, str] = None):
    """
    Writes a list of data objects to files in a directory.

    Args:
        directory_name: The name/path of the directory to write files to.
        file_data_ls: A list of data objects to write.
        data_type: The file format to use for writing data. Supported formats include JSON, CSV, Pickle, HDF5, INI,
                   YAML, and XML.
        file_prefix: An optional prefix to append to each file name. If not provided, file names will have no prefix.
    """
    fast_file_manager.bulk_write_directory(
        directory_name, file_data_ls, data_type, file_prefix=file_prefix)


def bulk_read_directory(directory_name: str, skip_unsupported_data_type: bool = False,
                        sort_by: Callable = None, sort_reverse=False,
                        file_prefix: Union[None, str] = None, include_file_names: bool = False) -> List[Any]:
    """
    Reads files from a directory. File names must be in the same style and format as bulk_write_directory.

    Args:
        directory_name: The name/path of the directory to read files from.
        skip_unsupported_data_type: If True, skips files with unsupported file types.
        sort_by: An optional callable taking a file name as input and returning a sorting key. If provided, files will
                 be sorted based on this function's return values.
        sort_reverse: If True, sorts files in reverse order according to the sort_by callable.
        file_prefix: An optional file name prefix for filtering files to read. If not provided, all files in the
                     directory will be considered.
        include_file_names: If True, returns a dictionary mapping file names to the read data objects. If False,
                            returns a list of the read data objects.

    Returns:
        List[Any]: A list of data objects read from the directory, or a dictionary mapping file names to data objects
                   if `include_file_names` is True.
    """
    return fast_file_manager.bulk_read_directory(directory_name, skip_unsupported_data_type=skip_unsupported_data_type,
                                                 sort_by=sort_by, sort_reverse=sort_reverse, file_prefix=file_prefix,
                                                 include_file_names=include_file_names)



from typing import Any, Union, List, Callable

from fastfs.global_instance import fast_file_manager


def write_yaml(file_name: str, data: Any):
    """
    Writes data to a YAML file.

    Args:
        file_name: The name/path of the file to write the YAML data to.
        data: The data to write as a YAML object.
    """
    fast_file_manager.write_yaml(file_name, data)


def read_yaml(file_name: str) -> Any:
    """
    Reads data from a YAML file.

    Args:
        file_name: The name/path of the YAML file to read from.

    Returns:
        Any: The data read from the YAML file.
    """
    return fast_file_manager.read_yaml(file_name)


def write_hdf5(file_name: str, data: Any):
    """
    Writes data to an HDF5 file.

    Args:
        file_name: The name/path of the file to write the HDF5 data to.
        data: The data to write as an HDF5 object.
    """
    fast_file_manager.write_hdf5(file_name, data)


def read_hdf5(file_name: str) -> Any:
    """
    Reads data from an HDF5 file.

    Args:
        file_name: The name/path of the HDF5 file to read from.

    Returns:
        Any: The data read from the HDF5 file.
    """
    return fast_file_manager.read_hdf5(file_name)


def write_dataframe(file_name: str, dataframe: 'pd.DataFrame', sep: str = ',', header: Union[bool, List[str]] = True, index: bool = True):
    """
    Writes a pandas dataframe to a CSV file. If the extension provided in 'file_name' is not '.csv' it will be changed to that.

    Args:
        file_name: The name/path of the file to write the data to.
        dataframe: The data to write as a pandas DataFrame.
        sep (optional): The delimiter character for the csv output file.
        header (optional): Write out the column names. If a list of strings is given it is assumed to be aliases for the column names.
        index (optional): Write row names (index).
    """

    fast_file_manager.write_dataframe(
        file_name, dataframe, sep=sep, header=header, index=index)


def read_dataframe(file_name: str, sep: str = ','):
    """
    Reads a pandas dataframe from a CSV, JSON, or PICKLE file.

    Args:
        file_name: The name/path of the file to read the data from.
        sep (optional): The delimiter character if the input file is a CSV.
    """

    return fast_file_manager.read_dataframe(file_name, sep=sep)



from fastfs.file_managers.abstract_file_manager import AbstractFileManager
from fastfs.decorators import safe_read, safe_write, path_replace

from typing import Any, List, Union

from fastfs.exceptions import InvalidFileDataError, CorruptFileError, FileNotFound, MissingDependencyError, FileWriteError, FileReadError, UnsupportedFileType


try:
    import yaml
except ImportError:
    yaml = None

try:
    import h5py
except ImportError:
    h5py = None


try:
    import pandas as pd
except ImportError:
    pd = None


class ExtensionFileManager(AbstractFileManager):

    @safe_write()
    def write_yaml(self, file, data: Any):
        try:
            if yaml is None:
                raise MissingDependencyError("PyYAML")

            yaml.dump(data, file)
        except yaml.YAMLError as exc:
            raise InvalidFileDataError('Failed to write YAML data.') from exc

    @safe_read()
    def read_yaml(self, file):
        try:
            if yaml is None:
                raise MissingDependencyError("PyYAML")

            return yaml.safe_load(file)
        except yaml.YAMLError as exc:
            raise CorruptFileError('Failed to read YAML data.') from exc

    @path_replace
    def write_hdf5(self, file_name: str, data: Any):
        try:
            if h5py is None:
                raise MissingDependencyError("hyp5")

            with h5py.File(file_name, 'w') as f:
                f.create_dataset('data', data=data)
        except (OSError, PermissionError, IsADirectoryError, InvalidFileDataError) as exc:
            raise FileWriteError from exc

    @path_replace
    def read_hdf5(self, file_name: str):
        try:
            if h5py is None:
                raise MissingDependencyError("hyp5")

            with h5py.File(file_name, 'r') as f:
                return f['data'][()]
        except FileNotFoundError as exc:
            raise FileNotFound(file_name) from exc
        except (OSError, PermissionError, IsADirectoryError) as exc:
            raise FileReadError from exc

    @path_replace
    def write_dataframe(self, file_name: str, dataframe: 'pd.DataFrame', sep: str = ',', header: Union[bool, List[str]] = True, index: bool = True):
        try:
            if pd is None:
                raise MissingDependencyError("pandas")

            file_extension = self.get_file_extension(file_name)

            if file_extension != '.csv':
                file_name = file_name.replace(file_extension, '.csv')

            dataframe.to_csv(file_name, sep=sep, header=header, index=index)

        except (OSError, PermissionError, IsADirectoryError, InvalidFileDataError) as exc:
            raise FileWriteError from exc

    @path_replace
    def read_dataframe(self, file_name: str, sep: str = ',') -> 'pd.DataFrame':
        try:
            if pd is None:
                raise MissingDependencyError("pandas")

            file_extension = self.get_file_extension(file_name)

            if file_extension == '.csv':
                return pd.read_csv(file_name, sep=sep)
            elif file_extension in ('.pickle', '.pkl'):
                return pd.read_pickle(file_name)
            elif file_extension == '.json':
                return pd.read_json(file_name)
            else:
                raise UnsupportedFileType(file_extension)

        except FileNotFoundError as exc:
            raise FileNotFound(file_name) from exc
        except (OSError, PermissionError, IsADirectoryError) as exc:
            raise FileReadError from exc



from enum import Enum


class FileTypes(Enum):
    PICKLE = ('pickle', 'pkl')
    JSON = 'json'
    BINARY = ('bin', 'binary', 'dat')
    CSV = 'csv'
    HDF5 = 'hdf5'
    INI = 'ini'
    YAML = ('yaml', 'yml')



import os
import unittest
from typing import List
import shutil
from fastfs.utils import sorted_ls, ls, file_exists, touch_file, touch_directory, delete_file, delete_directory, create_fs

class TestFastFsUtils(unittest.TestCase):

    def setUp(self):
        # Setup a test directory in the current directory
        self.test_dir = os.path.join(os.getcwd(), 'test_dir')

        os.mkdir(self.test_dir)

        self.test_file = 'test_file.txt'

    def tearDown(self):
        # Delete the test directory after running the tests
        shutil.rmtree(self.test_dir)

    def test_touch_directory(self):
        new_dir = os.path.join(self.test_dir, 'child_test_dir')

        touch_directory(new_dir)

        self.assertTrue(os.path.exists(new_dir))

    def test_delete_directory(self):
        new_dir = os.path.join(self.test_dir, 'child_test_dir')
        touch_directory(new_dir)

        self.assertTrue(os.path.exists(new_dir))

        delete_directory(new_dir)

        self.assertFalse(os.path.exists(new_dir))


    def test_file_exists(self):
        touch_file(self.test_file)
        self.assertTrue(file_exists(self.test_file))

    def test_ls(self):
        touch_file(os.path.join(self.test_dir, 'test_file.txt'))
        files: List[str] = ls(self.test_dir)
        self.assertIn('test_file.txt', files)  # Check if the created file is in the directory
    
    def test_sorted_ls(self):
        test_files = ['3.pkl', '1.pkl', '2.pkl']

        for file_name in test_files:
            with open(f"{self.test_dir}/{file_name}", "w") as f:
                f.write('')

        # Call the function
        sorted_files = sorted_ls(
            self.test_dir,
            lambda file_name: int(file_name.replace('.pkl', '')),
            reverse=False
        )

        # Assert that the files are sorted as expected
        self.assertEqual(sorted_files, ['1.pkl', '2.pkl', '3.pkl'])

        # Call the function with reverse
        reverse_sorted_files = sorted_ls(
            self.test_dir,
            lambda file_name: int(file_name.replace('.pkl', '')),
            reverse=True
        )

        # Assert that the files are reverse sorted as expected
        self.assertEqual(reverse_sorted_files, ['3.pkl', '2.pkl', '1.pkl'])

    def test_delete_file(self):
        touch_file(self.test_file)
        delete_file(self.test_file)
        self.assertFalse(file_exists(self.test_file))  # Check if file was deleted

if __name__ == '__main__':
    unittest.main()


import os
import unittest
from typing import List
import shutil
from fastfs.extensions import write_dataframe, read_dataframe, write_hdf5, read_hdf5, write_yaml, read_yaml


try:
    import yaml
except ImportError:
    yaml = None

try:
    import h5py
except ImportError:
    h5py = None


try:
    import pandas as pd
except ImportError:
    pd = None


class TestFastFsExtensions(unittest.TestCase):

    def setUp(self):
        # Setup a test directory in the current directory
        self.test_dir = os.path.join(os.getcwd(), 'test_dir')

        os.mkdir(self.test_dir)

        self.df_path = os.path.join(self.test_dir, 'pd_test.csv')
        self.h5_path = os.path.join(self.test_dir, 'test.hdf5')
        self.yaml_path = os.path.join(self.test_dir, 'test.yml')

    def tearDown(self):
        # Delete the test directory after running the tests
        shutil.rmtree(self.test_dir)

    def test_dataframe_file_write_and_read(self):

        if pd is None:
            self.skipTest(
                'pandas optional dependency is not installed. Skipping test...')

        test_data = {'col1': ['test', '456']}

        df = pd.DataFrame.from_dict(test_data)
        df.columns = df.columns.astype(str)

        write_dataframe(self.df_path, df, index=False)

        file_df = read_dataframe(self.df_path)
        file_df.columns = file_df.columns.astype(str)

        # Convert dataframes to dictionary and compare
        self.assertDictEqual(df.to_dict(), file_df.to_dict())

    def test_hdf5_file_write_and_read(self):

        if h5py is None:
            self.skipTest(
                'h5py optional dependency is not installed. Skipping test...')

        # Create a simple list as the test data
        data = [1, 2, 3, 4, 5]

        # Write the data to a HDF5 file
        write_hdf5(self.h5_path, data)

        # Read the data back from the HDF5 file
        file_data = read_hdf5(self.h5_path)

        # Compare the original data and the data read from file
        self.assertListEqual(data, list(file_data))

    def test_yaml_file_write_and_read(self):

        if yaml is None:
            self.skipTest(
                'PyYAML optional dependency is not installed. Skipping test...')

        # Create a dictionary as test data
        data = {'key1': 'value1', 'key2': 'value2', 'key3': 'value3'}

        # Write the data to a YAML file
        write_yaml(self.yaml_path, data)

        # Read the data back from the YAML file
        file_data = read_yaml(self.yaml_path)

        # Compare the original data and the data read from file
        self.assertDictEqual(data, file_data)


if __name__ == '__main__':
    unittest.main()



from fastfs.file_managers.base_file_manager import BaseFileExtensionManager
from fastfs.data_types import FileTypes
from fastfs.decorators import safe_read, safe_write, path_replace

from fastfs.exceptions import DirectoryNotFound, BulkReadDirectoryError, UnsupportedFileType

from typing import Any, List, Union, Callable

import json
import os
import time


class AbstractFileManager(BaseFileExtensionManager):

    @path_replace
    def get_file_extension(self, file_name):
        return os.path.splitext(file_name)[1]

    @path_replace
    def get_directory_info(self, directory_name):
        info = {}

        # Check if directory exists
        if os.path.isdir(directory_name):
            # Get the absolute path of the directory
            info["absolute_path"] = os.path.abspath(directory_name)

            # Get the number of files in the directory
            info["num_files"] = len(os.listdir(directory_name))

            # Get the creation time of the directory
            creation_time = os.path.getctime(directory_name)
            info["creation_time"] = time.ctime(creation_time)

            # Get the modification time of the directory
            modification_time = os.path.getmtime(directory_name)
            info["modification_time"] = time.ctime(modification_time)

            # Get the size of the directory
            total_size = 0
            for dirpath, dirnames, filenames in os.walk(directory_name):
                for f in filenames:
                    fp = os.path.join(dirpath, f)
                    if not os.path.islink(fp):
                        total_size += os.path.getsize(fp)
            info["total_size"] = total_size

        else:
            raise DirectoryNotFound(directory_name)

        return info

    @path_replace
    def get_file_info(self, file_name):
        info = {}

        # Check if file exists
        if os.path.isfile(file_name):
            # Get the absolute path of the file
            info["absolute_path"] = os.path.abspath(file_name)

            # Get the creation time of the file
            creation_time = os.path.getctime(file_name)
            info["creation_time"] = time.ctime(creation_time)

            # Get the modification time of the file
            modification_time = os.path.getmtime(file_name)
            info["modification_time"] = time.ctime(modification_time)

            # Get the size of the file
            info["size"] = os.path.getsize(file_name)

            # Get file extension
            info["extension"] = self.get_file_extension(file_name)

        else:
            raise ValueError(f"{file_name} does not exist")

        return info

    @safe_write()
    def write_lines(self, file, lines: list):

        for line in lines:
            file.write(line + os.linesep)

    @safe_read()
    def read_lines(self, file):
        lines = [line.strip() for line in file.readlines()]

        return lines

    @safe_read(context_manager=False)
    def iter_lines(self, file):
        for line in file:
            yield line

        file.close()

    @path_replace
    def bulk_write_directory(self, directory_name: str, file_data_ls: List[Any], data_type: Union[FileTypes, str],
                             file_prefix: Union[None, str] = None):

        if isinstance(data_type, str):
            try:
                data_type = FileTypes[data_type.upper()]
            except KeyError as exc:
                raise UnsupportedFileType(data_type) from exc

        if data_type not in FileTypes:
            supported_types = ", ".join(FileTypes.__members__.keys())
            raise UnsupportedFileType(
                data_type, supported_types=supported_types)

        # First, create the directory if it doesn't exist
        self.touch_directory(directory_name)

        file_extension = data_type.value

        # Then, iterate over the file data and write each file
        for idx, file_data in enumerate(file_data_ls):

            full_path = f'{directory_name}/{idx}'

            if file_prefix is not None:
                full_path += f'-{file_prefix}'

            full_path += f'.{file_extension}'

            if data_type == FileTypes.JSON:
                self.write_json(full_path, file_data)
            elif data_type == FileTypes.PICKLE:
                self.write_pickle(full_path, file_data)
            elif data_type == FileTypes.CSV:
                self.write_csv(full_path, file_data)
            elif data_type == FileTypes.BINARY:
                self.write_binary(full_path, file_data)

    @path_replace
    def bulk_read_directory(self, directory_name: str, skip_unsupported_data_type: bool = False,
                            sort_by: Callable = None, sort_reverse=False,
                            file_prefix: Union[None, str] = None, include_file_names: bool = False) -> List[Any]:
        data = {}

        if sort_by == None:

            if file_prefix == None:
                def sort_by(file_name): return int(
                    file_name.replace(self.get_file_extension(file_name), ''))
            else:
                def sort_by(file_name): return int(file_name.split("-")[0])

        sorted_file_names = self.sorted_ls(
            directory_name, sort_by=sort_by, reverse=sort_reverse)

        if len(set(sorted_file_names)) != len(sorted_file_names):
            raise BulkReadDirectoryError(
                'Duplicate files found. File names should be unique.')

        for file_name in sorted_file_names:

            file_extension = self.get_file_extension(
                file_name).replace('.', '')

            if file_extension == 'json':
                data[file_name] = self.read_json(
                    f"{directory_name}/{file_name}")
            elif file_extension == 'pickle':
                data[file_name] = self.read_pickle(
                    f"{directory_name}/{file_name}")
            # ...add other data types as needed...
            else:

                if skip_unsupported_data_type:
                    continue

                raise UnsupportedFileType(f'.{file_extension}')

        if include_file_names:
            return data
        else:
            return list(data.values())



import os

from typing import Any, Callable, Union, List, Dict, Tuple

import shutil
import configparser

import json
import pickle
import csv


from fastfs.exceptions import FileWriteError, FileReadError, FileNotFound, InvalidFileDataError, CorruptFileError
from fastfs.decorators import path_replace, safe_read, safe_write


class BaseFileManager():
    def __init__(self):
        self._local_fs = None

        self._fs_active = False

    def _read_local_fs(self):
        if not os.path.exists('.fastfs'):
            return None, False

        fast_fs_config = self.read_ini('.fastfs')

        local_fs = fast_fs_config['fastfs_directory']
        fs_active = fast_fs_config['fastfs_active']

        return local_fs, fs_active

    @property
    def local_fs(self):
        if self._local_fs is None:
            self._local_fs, self._fs_active = self._read_local_fs()
        return self._local_fs

    def _path_replace(self, file_path: str) -> str:

        # If fastfs
        if not self._fs_active:
            return file_path

        if file_path == '.fastfs':
            return file_path

        if self.local_fs is None:
            return file_path

        if os.path.isabs(file_path):
            return file_path

        if not file_path.startswith(os.path.join('.', self.local_fs)):
            return os.path.join('.', self.local_fs, file_path)

        return file_path

    def create_fs(self, directory_name: str, active: bool = True):

        self.touch_directory(directory_name)

        config = {'directory': directory_name, 'active': active}

        self.write_ini('.fastfs', config)
        self._local_fs = directory_name
        self._fs_active = active

    @path_replace
    def _safe_write_func(self, file_name: str, func: Callable, file_data: Any,
                         write_mode='w', encoding='utf-8', *args, **kwargs):

        try:

            encoding = None if 'b' in write_mode else encoding

            # Open the file in write mode
            with open(file_name, write_mode, encoding=encoding) as file:
                # Call the decorated function
                func(self, file, file_data, *args, **kwargs)
        except (OSError, PermissionError, IsADirectoryError, InvalidFileDataError) as exc:
            raise FileWriteError from exc

    @path_replace
    def _safe_read_func(self, file_name: str, func: Callable, read_mode='r',
                        context_manager=True, encoding='utf-8', *args, **kwargs):

        try:

            encoding = None if 'b' in read_mode else encoding

            if context_manager:
                with open(file_name, read_mode, encoding=encoding) as file:
                    # Call the decorated function
                    return func(self, file, *args, **kwargs)

            else:
                file = open(file_name, read_mode, encoding=encoding)

                return func(self, file, *args, **kwargs)

        except FileNotFoundError as exc:
            raise FileNotFound(file_name) from exc
        except (OSError, PermissionError, IsADirectoryError) as exc:
            raise FileReadError from exc

    @path_replace
    def is_hidden_file(self, file_name):
        name = os.path.basename(os.path.abspath(file_name))
        return name.startswith('.')

    @path_replace
    def touch_directory(self, directory_name: str):

        if not self.file_exists(directory_name):
            os.mkdir(directory_name)

    @path_replace
    def file_exists(self, file_name: str) -> bool:

        return os.path.exists(file_name)

    def get_fs_directory(self, absolute_path=False):
        if self.file_exists('.fastfs'):
            config = self.read_ini('.fastfs')

            path = config['directory']

            if absolute_path:
                path = os.path.abspath(path)

            return path

        return None

    @path_replace
    def delete_file(self, file_name: str):
        os.remove(file_name)

    @path_replace
    def delete_directory(self, directory_name: str):
        shutil.rmtree(directory_name)

    @path_replace
    def ls(self, directory: str, show_hidden=False):

        files = os.listdir(directory)

        if show_hidden:
            return files
        else:
            return [file for file in files if not self.is_hidden_file(file)]

    def _remove_extension(self, file_name):
        root, ext = os.path.splitext(file_name)
        if ext == '':
            return file_name
        return root

    def sorted_ls(self, directory: str, sort_by: Callable[[str], Any], reverse: bool = False) -> List[str]:

        ls = self.ls(directory)

        return sorted(ls, key=sort_by, reverse=reverse)

    @safe_write()
    def write_file(self, file, file_data):
        if not isinstance(file_data, str):
            file_data = str(file_data)

        file.write(file_data)

    @safe_read()
    def read_file(self, file):
        return file.read()

    @path_replace
    def touch_file(self, file_name: str):
        if not self.file_exists(file_name):
            self.write_file(file_name, '')

    # We only include read/write ini functions in this class
    # so that we can use it for .fastfs config

    @safe_write()
    def write_ini(self, file, data: Union[Dict[Any, Dict[Any, Any]], Dict[Any, Any]]):

        # not a dict or is an empty dict
        if not isinstance(data, dict) or not data:
            raise InvalidFileDataError(
                'INI input data must either be a dict or a dict of dicts.')

        first_key = next(iter(data))

        # If it's a flat dict, nest it in another dict under 'DEFAULT' key
        if not isinstance(data[first_key], dict):
            data = {'DEFAULT': data}

        try:
            config = configparser.ConfigParser()
            config.read_dict(data)
        except configparser.DuplicateSectionError as exc:
            raise InvalidFileDataError(
                'Duplicate section found in data.') from exc
        except configparser.DuplicateOptionError as exc:
            raise InvalidFileDataError(
                'Duplicate option in selection found.') from exc
        except ValueError as exc:
            raise InvalidFileDataError(
                'For .INI files, keys are section names, values are dictionaries with keys and values that should be present in the section.') from exc

        config.write(file)

    @safe_read()
    def read_ini(self, file) -> Union[Dict[Any, Dict[Any, Any]], Dict[Any, Any]]:
        config = configparser.ConfigParser()

        try:
            config.read_file(file)
        except configparser.ParsingError as exc:
            raise InvalidFileDataError(
                'The file could not be parsed.') from exc
        except configparser.DuplicateSectionError as exc:
            raise InvalidFileDataError(
                'Duplicate section found in the file.') from exc
        except configparser.DuplicateOptionError as exc:
            raise InvalidFileDataError(
                'Duplicate option in a section found in the file.') from exc

        # DEFAULT not included in config.sections()
        data_dict = {'DEFAULT': dict(config['DEFAULT'])}

        for section in config.sections():
            data_dict[section] = dict(config.items(section))

        # Return flat dictionary if there's only 'DEFAULT' section
        if len(data_dict.keys()) == 1 and 'DEFAULT' in data_dict:
            return data_dict['DEFAULT']

        return data_dict


class BaseFileExtensionManager(BaseFileManager):

    @safe_write(write_mode='wb')
    def write_binary(self, file, file_data: bytes):
        if not isinstance(file_data, bytes):
            raise ValueError("Data should be bytes for writing binary.")
        file.write(file_data)

    @safe_read(read_mode='rb')
    def read_binary(self, file):
        return file.read()

    @safe_write(write_mode='wb')
    def write_pickle(self, file, file_data):

        try:
            pickle.dump(file_data, file)
        except pickle.PicklingError as exc:
            raise InvalidFileDataError(
                'Object may not be serializable.') from exc
        except (OverflowError, MemoryError, AttributeError) as exc:
            raise FileWriteError from exc

    @safe_read(read_mode='rb')
    def read_pickle(self, file):

        try:
            return pickle.load(file)
        except pickle.UnpicklingError as exc:
            raise CorruptFileError('Failed to unpickle the file.') from exc
        except EOFError as exc:
            raise CorruptFileError(
                'Unexpected end of file while reading pickle data.') from exc
        except AttributeError as exc:
            raise CorruptFileError(
                'A required attribute was not found.') from exc
        except ImportError as exc:
            raise MissingDependencyError(
                '(see above exception)') from exc
        except pickle.PickleError as exc:
            raise CorruptFileError(
                'An error occured while unpickling.') from exc

    @safe_write()
    def write_json(self, file, file_data):

        try:
            json.dump(file_data, file)
        except TypeError as exc:
            raise InvalidFileDataError(
                'Failed to serialize the data.') from exc
        except OverflowError as exc:
            raise InvalidFileDataError(
                'The data is too deeply nested.') from exc
        except json.JSONDecodeError as exc:
            raise InvalidFileDataError('Could not encode JSON.') from exc

    @safe_read()
    def read_json(self, file):
        try:
            return json.load(file)
        except json.JSONDecodeError as exc:
            raise InvalidFileDataError('Could not decode JSON.') from exc

    @safe_write()
    def write_csv(self, file, file_data: Union[List[dict], List[list]], header: Union[None, list] = None):

        first_item = file_data[0]

        try:

            if isinstance(first_item, dict):
                field_names = first_item.keys()
                writer = csv.DictWriter(file, fieldnames=field_names)
                writer.writeheader()
            else:
                if header is None:
                    raise InvalidFileDataError(
                        "A header is required when using a list of lists for CSV file data.")

                writer = csv.writer(file)
                writer.writerow(header)

            # Write the actual data
            writer.writerows(file_data)
        except csv.Error as exc:
            raise InvalidFileDataError('Failed to write CSV data.') from exc
        except (AttributeError, TypeError) as exc:
            raise InvalidFileDataError('The data is not writable.') from exc

    @safe_read()
    def read_csv(self, file, return_list_of_dicts: bool = False) -> Union[Tuple[List[str], List[List[str]]], List[dict]]:
        try:
            if return_list_of_dicts:
                reader = csv.DictReader(file)
                return list(reader)
            else:
                reader = csv.reader(file)
                headers = next(reader)  # First line should be headers
                data = list(reader)
                return headers, data

        except csv.Error as exc:
            raise InvalidFileDataError('Failed to read CSV data.') from exc
        except (AttributeError, TypeError) as exc:
            raise InvalidFileDataError('The data is not readable.') from exc
